import random
from abc import ABC, abstractproperty
from eolib.data.eo_numeric_limits import CHAR_MAX


class SequenceStart(ABC):
    """
    A value sent by the server to update the client's sequence start, also known as the 'starting
    counter ID'.
    """

    @abstractproperty
    def value(self) -> int:
        """
        int: Gets the sequence start value.
        """
        raise NotImplementedError()

    @staticmethod
    def zero() -> 'SequenceStart':
        """
        Returns an instance of SequenceStart with a value of 0.

        Returns:
            SequenceStart: An instance of SequenceStart.
        """
        return SimpleSequenceStart(0)


class SimpleSequenceStart(SequenceStart):
    _value: int

    def __init__(self, value: int):
        self._value = value

    @property
    def value(self):
        return self._value


class AccountReplySequenceStart(SimpleSequenceStart):
    """
    A class representing the sequence start value sent with the `ACCOUNT_REPLY` server packet.

    See Also:
        - [`AccountReplyServerPacket`][eolib.protocol._generated.net.server.AccountReplyServerPacket]
    """

    def __init__(self, value: int):
        super().__init__(value)

    @staticmethod
    def from_value(value: int) -> 'AccountReplySequenceStart':
        """
        Creates an instance of AccountReplySequenceStart from the value sent with the
        `ACCOUNT_REPLY` server packet.

        Args:
            value (int): The sequence start value sent with the `ACCOUNT_REPLY` server packet.

        Returns:
            AccountReplySequenceStart: An instance of AccountReplySequenceStart.
        """
        return AccountReplySequenceStart(value)

    @staticmethod
    def generate() -> 'AccountReplySequenceStart':
        """
        Generates an instance of AccountReplySequenceStart with a random value in the range 0-240.

        Returns:
            AccountReplySequenceStart: An instance of AccountReplySequenceStart.
        """
        return AccountReplySequenceStart(random.randrange(0, 240))


class InitSequenceStart(SimpleSequenceStart):
    """
    A class representing the sequence start value sent with the `INIT_INIT` server packet.

    See Also:
        - [`InitInitServerPacket`][eolib.protocol._generated.net.server.InitInitServerPacket]
    """

    _seq1: int
    _seq2: int

    def __init__(self, value: int, seq1: int, seq2: int):
        super().__init__(value)
        self._seq1 = seq1
        self._seq2 = seq2

    @property
    def seq1(self) -> int:
        """
        Returns the `seq1` byte value sent with the `INIT_INIT` server packet.

        Returns:
            int: The seq1 byte value.
        """
        return self._seq1

    @property
    def seq2(self) -> int:
        """
        Returns the `seq2` byte value sent with the `INIT_INIT` server packet.

        Returns:
            int: The seq2 byte value.
        """
        return self._seq2

    @staticmethod
    def from_init_values(seq1: int, seq2: int) -> 'InitSequenceStart':
        """
        Creates an instance of InitSequenceStart from the values sent with the `INIT_INIT` server
        packet.

        Args:
            seq1 (int): The `seq1` byte value sent with the `INIT_INIT` server packet.
            seq2 (int): The `seq2` byte value sent with the `INIT_INIT` server packet.

        Returns:
            InitSequenceStart: An instance of InitSequenceStart
        """
        value = seq1 * 7 + seq2 - 13
        return InitSequenceStart(value, seq1, seq2)

    @staticmethod
    def generate() -> 'InitSequenceStart':
        """
        Generates an instance of InitSequenceStart with a random value in the range 0-1757.

        Returns:
            InitSequenceStart: An instance of InitSequenceStart.
        """
        value = random.randrange(0, 1757)
        seq1_max = int((value + 13) / 7)
        seq1_min = max(0, int((value - (CHAR_MAX - 1) + 13 + 6) / 7))

        seq1 = random.randrange(0, seq1_max - seq1_min) + seq1_min
        seq2 = value - seq1 * 7 + 13

        return InitSequenceStart(value, seq1, seq2)


class PingSequenceStart(SimpleSequenceStart):
    """
    A class representing the sequence start value sent with the `CONNECTION_PLAYER` server packet.

    See Also:
        - [`ConnectionPlayerServerPacket`][eolib.protocol._generated.net.server.ConnectionPlayerServerPacket]
    """

    def __init__(self, value: int, seq1: int, seq2: int):
        super().__init__(value)
        self._seq1 = seq1
        self._seq2 = seq2

    @property
    def seq1(self) -> int:
        """
        Returns the seq1 short value sent with the `CONNECTION_PLAYER` server packet.

        Returns:
            int: The seq1 short value.
        """
        return self._seq1

    @property
    def seq2(self) -> int:
        """
        Returns the seq2 char value sent with the `CONNECTION_PLAYER` server packet.

        Returns:
            int: The seq2 char value.
        """
        return self._seq2

    @staticmethod
    def from_ping_values(seq1: int, seq2: int) -> 'PingSequenceStart':
        """
        Creates an instance of PingSequenceStart from the values sent with the `CONNECTION_PLAYER`
        server packet.

        Args:
            seq1 (int): The `seq1` short value sent with the `CONNECTION_PLAYER` server packet.
            seq2 (int): The `seq2` char value sent with the `CONNECTION_PLAYER` server packet.

        Returns:
            PingSequenceStart: An instance of PingSequenceStart.
        """
        value = seq1 - seq2
        return PingSequenceStart(value, seq1, seq2)

    @staticmethod
    def generate() -> 'PingSequenceStart':
        """
        Generates an instance of PingSequenceStart with a random value in the range 0-1757.

        Returns:
            PingSequenceStart: An instance of PingSequenceStart.
        """
        value = random.randrange(0, 1757)
        seq1 = value + random.randrange(0, CHAR_MAX - 1)
        seq2 = seq1 - value

        return PingSequenceStart(value, seq1, seq2)


__all__ = ['SequenceStart', 'AccountReplySequenceStart', 'InitSequenceStart', 'PingSequenceStart']



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from ....serialization_error import SerializationError
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class BoardPostListing:
    """
    An entry in the list of town board posts
    """
    _byte_size: int = 0
    _post_id: int
    _author: str
    _subject: str

    def __init__(self, *, post_id: int, author: str, subject: str):
        """
        Create a new instance of BoardPostListing.

        Args:
            post_id (int): (Value range is 0-64008.)
            author (str): 
            subject (str): 
        """
        self._post_id = post_id
        self._author = author
        self._subject = subject

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def post_id(self) -> int:
        return self._post_id

    @property
    def author(self) -> str:
        return self._author

    @property
    def subject(self) -> str:
        return self._subject

    @staticmethod
    def serialize(writer: EoWriter, data: "BoardPostListing") -> None:
        """
        Serializes an instance of `BoardPostListing` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (BoardPostListing): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            writer.string_sanitization_mode = True
            if data._post_id is None:
                raise SerializationError("post_id must be provided.")
            writer.add_short(data._post_id)
            writer.add_byte(0xFF)
            if data._author is None:
                raise SerializationError("author must be provided.")
            writer.add_string(data._author)
            writer.add_byte(0xFF)
            if data._subject is None:
                raise SerializationError("subject must be provided.")
            writer.add_string(data._subject)
            writer.string_sanitization_mode = False
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "BoardPostListing":
        """
        Deserializes an instance of `BoardPostListing` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            BoardPostListing: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            reader.chunked_reading_mode = True
            post_id = reader.get_short()
            reader.next_chunk()
            author = reader.get_string()
            reader.next_chunk()
            subject = reader.get_string()
            reader.chunked_reading_mode = False
            result = BoardPostListing(post_id=post_id, author=author, subject=subject)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"BoardPostListing(byte_size={repr(self._byte_size)}, post_id={repr(self._post_id)}, author={repr(self._author)}, subject={repr(self._subject)})"



from eolib.data.eo_numeric_limits import CHAR_MAX, SHORT_MAX, THREE_MAX


def encode_number(number: int) -> bytes:
    """
    Encodes a number to a sequence of bytes.

    Args:
        number (int): The number to encode.

    Returns:
        bytes: The encoded sequence of bytes.
    """
    value = number
    d = 0xFE
    if number >= THREE_MAX:
        d = value // THREE_MAX + 1
        value = value % THREE_MAX

    c = 0xFE
    if number >= SHORT_MAX:
        c = value // SHORT_MAX + 1
        value = value % SHORT_MAX

    b = 0xFE
    if number >= CHAR_MAX:
        b = value // CHAR_MAX + 1
        value = value % CHAR_MAX

    a = value + 1

    return bytes([a, b, c, d])


def decode_number(encoded_number: bytes) -> int:
    """
    Decodes a number from a sequence of bytes.

    Args:
        encoded_number (bytes): The sequence of bytes to decode.

    Returns:
        int: The decoded number.
    """
    result = 0
    length = min(len(encoded_number), 4)

    for i in range(length):
        value = encoded_number[i]

        if value == 0xFE:
            break

        value -= 1

        if i == 0:
            result += value
        elif i == 1:
            result += CHAR_MAX * value
        elif i == 2:
            result += SHORT_MAX * value
        elif i == 3:
            result += THREE_MAX * value

    return result



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from __future__ import annotations
from typing import Union
from collections.abc import Iterable
from .guild_info_type import GuildInfoType
from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class GuildAgreeClientPacket(Packet):
    """
    Update the guild description or rank list
    """
    _byte_size: int = 0
    _session_id: int
    _info_type: GuildInfoType
    _info_type_data: 'GuildAgreeClientPacket.InfoTypeData'

    def __init__(self, *, session_id: int, info_type: GuildInfoType, info_type_data: 'GuildAgreeClientPacket.InfoTypeData' = None):
        """
        Create a new instance of GuildAgreeClientPacket.

        Args:
            session_id (int): (Value range is 0-4097152080.)
            info_type (GuildInfoType): 
            info_type_data (GuildAgreeClientPacket.InfoTypeData): Data associated with the `info_type` field.
        """
        self._session_id = session_id
        self._info_type = info_type
        self._info_type_data = info_type_data

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def session_id(self) -> int:
        return self._session_id

    @property
    def info_type(self) -> GuildInfoType:
        return self._info_type

    @property
    def info_type_data(self) -> 'GuildAgreeClientPacket.InfoTypeData':
        """
        GuildAgreeClientPacket.InfoTypeData: Data associated with the `info_type` field.
        """
        return self._info_type_data

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.Guild

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Agree

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        GuildAgreeClientPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "GuildAgreeClientPacket") -> None:
        """
        Serializes an instance of `GuildAgreeClientPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (GuildAgreeClientPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            writer.string_sanitization_mode = True
            if data._session_id is None:
                raise SerializationError("session_id must be provided.")
            writer.add_int(data._session_id)
            if data._info_type is None:
                raise SerializationError("info_type must be provided.")
            writer.add_short(int(data._info_type))
            if data._info_type == GuildInfoType.Description:
                if not isinstance(data._info_type_data, GuildAgreeClientPacket.InfoTypeDataDescription):
                    raise SerializationError("Expected info_type_data to be type GuildAgreeClientPacket.InfoTypeDataDescription for info_type " + GuildInfoType(data._info_type).name + ".")
                GuildAgreeClientPacket.InfoTypeDataDescription.serialize(writer, data._info_type_data)
            elif data._info_type == GuildInfoType.Ranks:
                if not isinstance(data._info_type_data, GuildAgreeClientPacket.InfoTypeDataRanks):
                    raise SerializationError("Expected info_type_data to be type GuildAgreeClientPacket.InfoTypeDataRanks for info_type " + GuildInfoType(data._info_type).name + ".")
                GuildAgreeClientPacket.InfoTypeDataRanks.serialize(writer, data._info_type_data)
            writer.string_sanitization_mode = False
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "GuildAgreeClientPacket":
        """
        Deserializes an instance of `GuildAgreeClientPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            GuildAgreeClientPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            reader.chunked_reading_mode = True
            session_id = reader.get_int()
            info_type = GuildInfoType(reader.get_short())
            info_type_data: GuildAgreeClientPacket.InfoTypeData = None
            if info_type == GuildInfoType.Description:
                info_type_data = GuildAgreeClientPacket.InfoTypeDataDescription.deserialize(reader)
            elif info_type == GuildInfoType.Ranks:
                info_type_data = GuildAgreeClientPacket.InfoTypeDataRanks.deserialize(reader)
            reader.chunked_reading_mode = False
            result = GuildAgreeClientPacket(session_id=session_id, info_type=info_type, info_type_data=info_type_data)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"GuildAgreeClientPacket(byte_size={repr(self._byte_size)}, session_id={repr(self._session_id)}, info_type={repr(self._info_type)}, info_type_data={repr(self._info_type_data)})"

    InfoTypeData = Union['GuildAgreeClientPacket.InfoTypeDataDescription', 'GuildAgreeClientPacket.InfoTypeDataRanks', None]
    """
    Data associated with different values of the `info_type` field.
    """

    class InfoTypeDataDescription:
        """
        Data associated with info_type value GuildInfoType.Description
        """
        _byte_size: int = 0
        _description: str

        def __init__(self, *, description: str):
            """
            Create a new instance of GuildAgreeClientPacket.InfoTypeDataDescription.

            Args:
                description (str): 
            """
            self._description = description

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size

        @property
        def description(self) -> str:
            return self._description

        @staticmethod
        def serialize(writer: EoWriter, data: "GuildAgreeClientPacket.InfoTypeDataDescription") -> None:
            """
            Serializes an instance of `GuildAgreeClientPacket.InfoTypeDataDescription` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (GuildAgreeClientPacket.InfoTypeDataDescription): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                if data._description is None:
                    raise SerializationError("description must be provided.")
                writer.add_string(data._description)
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "GuildAgreeClientPacket.InfoTypeDataDescription":
            """
            Deserializes an instance of `GuildAgreeClientPacket.InfoTypeDataDescription` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                GuildAgreeClientPacket.InfoTypeDataDescription: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                description = reader.get_string()
                result = GuildAgreeClientPacket.InfoTypeDataDescription(description=description)
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"GuildAgreeClientPacket.InfoTypeDataDescription(byte_size={repr(self._byte_size)}, description={repr(self._description)})"

    class InfoTypeDataRanks:
        """
        Data associated with info_type value GuildInfoType.Ranks
        """
        _byte_size: int = 0
        _ranks: tuple[str, ...]

        def __init__(self, *, ranks: Iterable[str]):
            """
            Create a new instance of GuildAgreeClientPacket.InfoTypeDataRanks.

            Args:
                ranks (Iterable[str]): (Length must be `9`.)
            """
            self._ranks = tuple(ranks)

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size

        @property
        def ranks(self) -> tuple[str, ...]:
            return self._ranks

        @staticmethod
        def serialize(writer: EoWriter, data: "GuildAgreeClientPacket.InfoTypeDataRanks") -> None:
            """
            Serializes an instance of `GuildAgreeClientPacket.InfoTypeDataRanks` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (GuildAgreeClientPacket.InfoTypeDataRanks): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                if data._ranks is None:
                    raise SerializationError("ranks must be provided.")
                if len(data._ranks) != 9:
                    raise SerializationError(f"Expected length of ranks to be exactly 9, got {len(data._ranks)}.")
                for i in range(9):
                    writer.add_string(data._ranks[i])
                    writer.add_byte(0xFF)
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "GuildAgreeClientPacket.InfoTypeDataRanks":
            """
            Deserializes an instance of `GuildAgreeClientPacket.InfoTypeDataRanks` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                GuildAgreeClientPacket.InfoTypeDataRanks: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                ranks = []
                for i in range(9):
                    ranks.append(reader.get_string())
                    reader.next_chunk()
                result = GuildAgreeClientPacket.InfoTypeDataRanks(ranks=ranks)
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"GuildAgreeClientPacket.InfoTypeDataRanks(byte_size={repr(self._byte_size)}, ranks={repr(self._ranks)})"



from protocol_code_generator.type.custom_type import CustomType


class StructType(CustomType):
    def __init__(self, name, fixed_size, bounded, source_path):
        self._name = name
        self._fixed_size = fixed_size
        self._bounded = bounded
        self._source_path = source_path

    @property
    def name(self):
        return self._name

    @property
    def fixed_size(self):
        return self._fixed_size

    @property
    def bounded(self):
        return self._bounded

    @property
    def source_path(self):
        return self._source_path



import pytest
from typing import Iterable
from eolib import EoReader


def test_slice():
    reader = create_reader([0x01, 0x02, 0x03, 0x04, 0x05, 0x06])
    reader.get_byte()
    reader.chunked_reading_mode = True

    reader2 = reader.slice()
    assert reader2.position == 0
    assert reader2.remaining == 5
    assert not reader2.chunked_reading_mode

    reader3 = reader2.slice(1)
    assert reader3.position == 0
    assert reader3.remaining == 4
    assert not reader3.chunked_reading_mode

    reader4 = reader3.slice(1, 2)
    assert reader4.position == 0
    assert reader4.remaining == 2
    assert not reader4.chunked_reading_mode

    assert reader.position == 1
    assert reader.remaining == 5
    assert reader.chunked_reading_mode


def test_slice_over_read():
    reader = create_reader([0x01, 0x02, 0x03])
    assert reader.slice(2, 5).remaining == 1
    assert reader.slice(3).remaining == 0
    assert reader.slice(4).remaining == 0
    assert reader.slice(4, 12345).remaining == 0


def test_slice_negative_index():
    reader = create_reader([0x01, 0x02, 0x03])
    with pytest.raises(ValueError):
        reader.slice(-1)


def test_slice_negative_length():
    reader = create_reader([0x01, 0x02, 0x03])
    with pytest.raises(ValueError):
        reader.slice(0, -1)


@pytest.mark.parametrize("byte_value", [0x00, 0x01, 0x02, 0x80, 0xFD, 0xFE, 0xFF])
def test_get_byte(byte_value: int):
    reader = create_reader([byte_value])
    assert reader.get_byte() == byte_value


def test_over_read_byte():
    reader = create_reader([])
    assert reader.get_byte() == 0x00


def test_get_bytes():
    reader = create_reader([0x01, 0x02, 0x03, 0x04, 0x05])
    assert reader.get_bytes(3) == bytes([0x01, 0x02, 0x03])
    assert reader.get_bytes(10) == bytes([0x04, 0x05])
    assert reader.get_bytes(1) == bytes([])


def test_get_char():
    reader = create_reader([0x01, 0x02, 0x80, 0x81, 0xFD, 0xFE, 0xFF])
    assert reader.get_char() == 0
    assert reader.get_char() == 1
    assert reader.get_char() == 127
    assert reader.get_char() == 128
    assert reader.get_char() == 252
    assert reader.get_char() == 0
    assert reader.get_char() == 254


def test_get_short():
    reader = create_reader(
        [0x01, 0xFE, 0x02, 0xFE, 0x80, 0xFE, 0xFD, 0xFE, 0xFE, 0xFE, 0xFE, 0x80, 0x7F, 0x7F, 0xFD]
        + [0xFD]
    )
    assert reader.get_short() == 0
    assert reader.get_short() == 1
    assert reader.get_short() == 127
    assert reader.get_short() == 252
    assert reader.get_short() == 0
    assert reader.get_short() == 0
    assert reader.get_short() == 32004
    assert reader.get_short() == 64008


def test_get_three():
    reader = create_reader(
        [0x01, 0xFE, 0xFE, 0x02, 0xFE, 0xFE, 0x80, 0xFE, 0xFE, 0xFD, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE]
        + [0xFE, 0x80, 0x81, 0x7F, 0x7F, 0xFE, 0xFD, 0xFD, 0xFE, 0xFD, 0xFD, 0xFD]
    )
    assert reader.get_three() == 0
    assert reader.get_three() == 1
    assert reader.get_three() == 127
    assert reader.get_three() == 252
    assert reader.get_three() == 0
    assert reader.get_three() == 0
    assert reader.get_three() == 32004
    assert reader.get_three() == 64008
    assert reader.get_three() == 16194276


def test_get_int():
    reader = create_reader(
        [0x01, 0xFE, 0xFE, 0xFE, 0x02, 0xFE, 0xFE, 0xFE, 0x80, 0xFE, 0xFE, 0xFE, 0xFD, 0xFE, 0xFE]
        + [0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0xFE, 0x80, 0x81, 0x82, 0x7F, 0x7F, 0xFE, 0xFE, 0xFD, 0xFD]
        + [0xFE, 0xFE, 0xFD, 0xFD, 0xFD, 0xFE, 0x7F, 0x7F, 0x7F, 0x7F, 0xFD, 0xFD, 0xFD, 0xFD]
    )
    assert reader.get_int() == 0
    assert reader.get_int() == 1
    assert reader.get_int() == 127
    assert reader.get_int() == 252
    assert reader.get_int() == 0
    assert reader.get_int() == 0
    assert reader.get_int() == 32004
    assert reader.get_int() == 64008
    assert reader.get_int() == 16194276
    assert reader.get_int() == 2048576040
    assert reader.get_int() == 4097152080


def test_get_string():
    reader = create_reader(b"Hello, World!")
    assert reader.get_string() == "Hello, World!"


def test_get_fixed_string():
    reader = create_reader(b"foobar")
    assert reader.get_fixed_string(3) == "foo"
    assert reader.get_fixed_string(3) == "bar"


def test_padded_get_fixed_string():
    reader = create_reader(bytes("fooÿbarÿÿÿ", "windows-1252"))
    assert reader.get_fixed_string(4, True) == "foo"
    assert reader.get_fixed_string(6, True) == "bar"


def test_chunked_get_string():
    reader = create_reader(bytes("Hello,ÿWorld!", "windows-1252"))
    reader.chunked_reading_mode = True

    assert reader.get_string() == "Hello,"

    reader.next_chunk()
    assert reader.get_string() == "World!"


def test_get_negative_length_string():
    reader = create_reader(b"foo")
    with pytest.raises(ValueError):
        reader.get_fixed_string(-1)


def test_get_encoded_string():
    reader = create_reader(b"!;a-^H s^3a:)")
    assert reader.get_encoded_string() == "Hello, World!"


def test_fixed_get_encoded_string():
    reader = create_reader(b"^0g[>k")
    assert reader.get_fixed_encoded_string(3) == "foo"
    assert reader.get_fixed_encoded_string(3) == "bar"


def test_padded_get_fixed_encoded_string():
    reader = create_reader(bytes("ÿ0^9ÿÿÿ-l=S>k", "windows-1252"))
    assert reader.get_fixed_encoded_string(4, True) == "foo"
    assert reader.get_fixed_encoded_string(6, True) == "bar"
    assert reader.get_fixed_encoded_string(3, True) == "baz"


def test_chunked_get_encoded_string():
    reader = create_reader(bytes("E0a3hWÿ!;a-^H", "windows-1252"))
    reader.chunked_reading_mode = True

    assert reader.get_encoded_string() == "Hello,"

    reader.next_chunk()
    assert reader.get_encoded_string() == "World!"


def test_get_negative_length_encoded_string():
    reader = create_reader(b"^0g")
    with pytest.raises(ValueError):
        reader.get_fixed_encoded_string(-1)


def test_chunked_reading_mode():
    reader = create_reader([])
    assert not reader.chunked_reading_mode
    reader.chunked_reading_mode = True
    assert reader.chunked_reading_mode


def test_remaining():
    reader = create_reader([0x01, 0x03, 0x04, 0xFE, 0x05, 0xFE, 0xFE, 0x06, 0xFE, 0xFE, 0xFE])

    assert reader.remaining == 11
    reader.get_byte()
    assert reader.remaining == 10
    reader.get_char()
    assert reader.remaining == 9
    reader.get_short()
    assert reader.remaining == 7
    reader.get_three()
    assert reader.remaining == 4
    reader.get_int()
    assert reader.remaining == 0

    reader.get_char()
    assert reader.remaining == 0


def test_chunked_remaining():
    reader = create_reader([0x01, 0x03, 0x04, 0xFF, 0x05, 0xFE, 0xFE, 0x06, 0xFE, 0xFE, 0xFE])
    reader.chunked_reading_mode = True

    assert reader.remaining == 3
    reader.get_char()
    reader.get_short()
    assert reader.remaining == 0

    reader.get_char()
    assert reader.remaining == 0

    reader.next_chunk()
    assert reader.remaining == 7


def test_next_chunk():
    reader = create_reader([0x01, 0x02, 0xFF, 0x03, 0x04, 0x5, 0xFF, 0x06])
    reader.chunked_reading_mode = True

    assert reader.position == 0

    reader.next_chunk()
    assert reader.position == 3

    reader.next_chunk()
    assert reader.position == 7

    reader.next_chunk()
    assert reader.position == 8

    reader.next_chunk()
    assert reader.position == 8


def test_next_chunk_not_in_chunked_reading_mode():
    reader = create_reader([0x01, 0x02, 0xFF, 0x03, 0x04, 0x5, 0xFF, 0x06])
    with pytest.raises(RuntimeError):
        reader.next_chunk()


def test_next_chunk_with_chunked_reading_toggled_in_between():
    reader = create_reader([0x01, 0x02, 0xFF, 0x03, 0x04, 0x5, 0xFF, 0x06])
    assert reader.position == 0

    reader.chunked_reading_mode = True
    reader.next_chunk()
    reader.chunked_reading_mode = False
    assert reader.position == 3

    reader.chunked_reading_mode = True
    reader.next_chunk()
    reader.chunked_reading_mode = False
    assert reader.position == 7

    reader.chunked_reading_mode = True
    reader.next_chunk()
    reader.chunked_reading_mode = False
    assert reader.position == 8

    reader.chunked_reading_mode = True
    reader.next_chunk()
    reader.chunked_reading_mode = False
    assert reader.position == 8


def test_under_read():
    reader = create_reader([0x7C, 0x67, 0x61, 0x72, 0x62, 0x61, 0x67, 0x65, 0xFF, 0xCA, 0x31])
    reader.chunked_reading_mode = True

    assert reader.get_char() == 123
    reader.next_chunk()
    assert reader.get_short() == 12345


def test_over_read():
    reader = create_reader([0xFF, 0x7C])
    reader.chunked_reading_mode = True

    assert reader.get_int() == 0
    reader.next_chunk()
    assert reader.get_short() == 123


def test_double_read():
    reader = create_reader([0xFF, 0x7C, 0xCA, 0x31])

    assert reader.get_int() == 790222478

    reader.chunked_reading_mode = True
    reader.next_chunk()
    assert reader.get_char() == 123
    assert reader.get_short() == 12345


def create_reader(input_data: Iterable[int]) -> EoReader:
    data = bytearray(10)
    data.extend(input_data)
    data.extend([0] * 10)

    return EoReader(data).slice(10, len(data) - 20)



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ...protocol_enum_meta import ProtocolEnumMeta

class MapTimedEffect(IntEnum, metaclass=ProtocolEnumMeta):
    """
    A timed effect that can occur on a map
    """
    None_ = 0
    HpDrain = 1
    TpDrain = 2
    Quake1 = 3
    Quake2 = 4
    Quake3 = 5
    Quake4 = 6



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ....protocol_enum_meta import ProtocolEnumMeta

class CharacterIcon(IntEnum, metaclass=ProtocolEnumMeta):
    """
    Icon displayed in paperdolls, books, and the online list
    """
    Player = 1
    Gm = 4
    Hgm = 5
    Party = 6
    GmParty = 9
    HgmParty = 10



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from ....serialization_error import SerializationError
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class GuildMember:
    """
    Information about a guild member
    """
    _byte_size: int = 0
    _rank: int
    _name: str
    _rank_name: str

    def __init__(self, *, rank: int, name: str, rank_name: str):
        """
        Create a new instance of GuildMember.

        Args:
            rank (int): (Value range is 0-252.)
            name (str): 
            rank_name (str): 
        """
        self._rank = rank
        self._name = name
        self._rank_name = rank_name

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def rank(self) -> int:
        return self._rank

    @property
    def name(self) -> str:
        return self._name

    @property
    def rank_name(self) -> str:
        return self._rank_name

    @staticmethod
    def serialize(writer: EoWriter, data: "GuildMember") -> None:
        """
        Serializes an instance of `GuildMember` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (GuildMember): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            writer.string_sanitization_mode = True
            if data._rank is None:
                raise SerializationError("rank must be provided.")
            writer.add_char(data._rank)
            writer.add_byte(0xFF)
            if data._name is None:
                raise SerializationError("name must be provided.")
            writer.add_string(data._name)
            writer.add_byte(0xFF)
            if data._rank_name is None:
                raise SerializationError("rank_name must be provided.")
            writer.add_string(data._rank_name)
            writer.string_sanitization_mode = False
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "GuildMember":
        """
        Deserializes an instance of `GuildMember` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            GuildMember: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            reader.chunked_reading_mode = True
            rank = reader.get_char()
            reader.next_chunk()
            name = reader.get_string()
            reader.next_chunk()
            rank_name = reader.get_string()
            reader.chunked_reading_mode = False
            result = GuildMember(rank=rank, name=name, rank_name=rank_name)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"GuildMember(byte_size={repr(self._byte_size)}, rank={repr(self._rank)}, name={repr(self._name)}, rank_name={repr(self._rank_name)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ...protocol_enum_meta import ProtocolEnumMeta

class ItemSpecial(IntEnum, metaclass=ProtocolEnumMeta):
    Normal = 0
    Rare = 1
    Legendary = 2
    Unique = 3
    Lore = 4
    Cursed = 5



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ...protocol_enum_meta import ProtocolEnumMeta

class SkillTargetType(IntEnum, metaclass=ProtocolEnumMeta):
    Normal = 0
    Self = 1
    Reserved2 = 2
    Group = 3



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from ....serialization_error import SerializationError
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class EquipmentChange:
    """
    Player equipment data.
    Sent when a player's visible equipment changes.
    Note that these values are graphic IDs.
    """
    _byte_size: int = 0
    _boots: int
    _armor: int
    _hat: int
    _weapon: int
    _shield: int

    def __init__(self, *, boots: int, armor: int, hat: int, weapon: int, shield: int):
        """
        Create a new instance of EquipmentChange.

        Args:
            boots (int): (Value range is 0-64008.)
            armor (int): (Value range is 0-64008.)
            hat (int): (Value range is 0-64008.)
            weapon (int): (Value range is 0-64008.)
            shield (int): (Value range is 0-64008.)
        """
        self._boots = boots
        self._armor = armor
        self._hat = hat
        self._weapon = weapon
        self._shield = shield

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def boots(self) -> int:
        return self._boots

    @property
    def armor(self) -> int:
        return self._armor

    @property
    def hat(self) -> int:
        return self._hat

    @property
    def weapon(self) -> int:
        return self._weapon

    @property
    def shield(self) -> int:
        return self._shield

    @staticmethod
    def serialize(writer: EoWriter, data: "EquipmentChange") -> None:
        """
        Serializes an instance of `EquipmentChange` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (EquipmentChange): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._boots is None:
                raise SerializationError("boots must be provided.")
            writer.add_short(data._boots)
            if data._armor is None:
                raise SerializationError("armor must be provided.")
            writer.add_short(data._armor)
            if data._hat is None:
                raise SerializationError("hat must be provided.")
            writer.add_short(data._hat)
            if data._weapon is None:
                raise SerializationError("weapon must be provided.")
            writer.add_short(data._weapon)
            if data._shield is None:
                raise SerializationError("shield must be provided.")
            writer.add_short(data._shield)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "EquipmentChange":
        """
        Deserializes an instance of `EquipmentChange` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            EquipmentChange: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            boots = reader.get_short()
            armor = reader.get_short()
            hat = reader.get_short()
            weapon = reader.get_short()
            shield = reader.get_short()
            result = EquipmentChange(boots=boots, armor=armor, hat=hat, weapon=weapon, shield=shield)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"EquipmentChange(byte_size={repr(self._byte_size)}, boots={repr(self._boots)}, armor={repr(self._armor)}, hat={repr(self._hat)}, weapon={repr(self._weapon)}, shield={repr(self._shield)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ...protocol_enum_meta import ProtocolEnumMeta

class PacketAction(IntEnum, metaclass=ProtocolEnumMeta):
    """
    The specific action that a packet performs.
    Part of the unique packet ID.
    """
    Request = 1
    Accept = 2
    Reply = 3
    Remove = 4
    Agree = 5
    Create = 6
    Add = 7
    Player = 8
    Take = 9
    Use = 10
    Buy = 11
    Sell = 12
    Open = 13
    Close = 14
    Msg = 15
    Spec = 16
    Admin = 17
    List = 18
    Tell = 20
    Report = 21
    Announce = 22
    Server = 23
    Drop = 24
    Junk = 25
    Obtain = 26
    Get = 27
    Kick = 28
    Rank = 29
    TargetSelf = 30
    TargetOther = 31
    TargetGroup = 33
    Dialog = 34
    Ping = 240
    Pong = 241
    Net242 = 242
    Net243 = 243
    Net244 = 244
    Error = 250
    Init = 255



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ....protocol_enum_meta import ProtocolEnumMeta

class MapDamageType(IntEnum, metaclass=ProtocolEnumMeta):
    """
    Type of damage being caused by the environment
    """
    TpDrain = 1
    Spikes = 2



import math
from eolib.packet.sequence_start import (
    SequenceStart,
    AccountReplySequenceStart,
    InitSequenceStart,
    PingSequenceStart,
)


def randrange_midpoint(a, b):
    return math.ceil((a + b) / 2)


def test_zero():
    assert SequenceStart.zero().value == 0


def test_account_reply_from_value():
    sequence_start = AccountReplySequenceStart.from_value(22)
    assert sequence_start.value == 22


def test_account_reply_generate(monkeypatch):
    monkeypatch.setattr('random.randrange', randrange_midpoint)

    sequence_start = AccountReplySequenceStart.generate()
    assert sequence_start.value == 120


def test_init__from_init_values():
    sequence_start = InitSequenceStart.from_init_values(110, 122)
    assert sequence_start.value == 879
    assert sequence_start.seq1 == 110
    assert sequence_start.seq2 == 122


def test_init_generate(monkeypatch):
    monkeypatch.setattr('random.randrange', randrange_midpoint)

    sequence_start = InitSequenceStart.generate()
    assert sequence_start.value == 879
    assert sequence_start.seq1 == 110
    assert sequence_start.seq2 == 122


def test_ping_from_ping_values():
    sequence_start = PingSequenceStart.from_ping_values(1005, 126)
    assert sequence_start.value == 879
    assert sequence_start.seq1 == 1005
    assert sequence_start.seq2 == 126


def test_ping_generate(monkeypatch):
    monkeypatch.setattr('random.randrange', randrange_midpoint)

    sequence_start = PingSequenceStart.generate()
    assert sequence_start.value == 879
    assert sequence_start.seq1 == 1005
    assert sequence_start.seq2 == 126



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

"""
Data structures generated from the [eo-protocol](https://github.com/cirras/eo-protocol/tree/master/xml/pub/server/protocol.xml){target="_blank"} XML specification.

Warning:
  - This subpackage should not be directly imported. 
  - Instead, import [eolib.protocol.pub.server][] (or the top-level `eolib`).
"""

from .talk_record import *
from .talk_message_record import *
from .talk_file import *
from .skill_master_skill_record import *
from .skill_master_record import *
from .skill_master_file import *
from .shop_trade_record import *
from .shop_record import *
from .shop_file import *
from .shop_craft_record import *
from .shop_craft_ingredient_record import *
from .inn_record import *
from .inn_question_record import *
from .inn_file import *
from .drop_record import *
from .drop_npc_record import *
from .drop_file import *



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from __future__ import annotations
from collections.abc import Iterable
from .guild_staff import GuildStaff
from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class GuildReportServerPacket(Packet):
    """
    Get guild info reply
    """
    _byte_size: int = 0
    _name: str
    _tag: str
    _create_date: str
    _description: str
    _wealth: str
    _ranks: tuple[str, ...]
    _staff_count: int
    _staff: tuple[GuildStaff, ...]

    def __init__(self, *, name: str, tag: str, create_date: str, description: str, wealth: str, ranks: Iterable[str], staff: Iterable[GuildStaff]):
        """
        Create a new instance of GuildReportServerPacket.

        Args:
            name (str): 
            tag (str): 
            create_date (str): 
            description (str): 
            wealth (str): 
            ranks (Iterable[str]): (Length must be `9`.)
            staff (Iterable[GuildStaff]): (Length must be 64008 or less.)
        """
        self._name = name
        self._tag = tag
        self._create_date = create_date
        self._description = description
        self._wealth = wealth
        self._ranks = tuple(ranks)
        self._staff = tuple(staff)
        self._staff_count = len(self._staff)

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def name(self) -> str:
        return self._name

    @property
    def tag(self) -> str:
        return self._tag

    @property
    def create_date(self) -> str:
        return self._create_date

    @property
    def description(self) -> str:
        return self._description

    @property
    def wealth(self) -> str:
        return self._wealth

    @property
    def ranks(self) -> tuple[str, ...]:
        return self._ranks

    @property
    def staff(self) -> tuple[GuildStaff, ...]:
        return self._staff

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.Guild

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Report

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        GuildReportServerPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "GuildReportServerPacket") -> None:
        """
        Serializes an instance of `GuildReportServerPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (GuildReportServerPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            writer.string_sanitization_mode = True
            if data._name is None:
                raise SerializationError("name must be provided.")
            writer.add_string(data._name)
            writer.add_byte(0xFF)
            if data._tag is None:
                raise SerializationError("tag must be provided.")
            writer.add_string(data._tag)
            writer.add_byte(0xFF)
            if data._create_date is None:
                raise SerializationError("create_date must be provided.")
            writer.add_string(data._create_date)
            writer.add_byte(0xFF)
            if data._description is None:
                raise SerializationError("description must be provided.")
            writer.add_string(data._description)
            writer.add_byte(0xFF)
            if data._wealth is None:
                raise SerializationError("wealth must be provided.")
            writer.add_string(data._wealth)
            writer.add_byte(0xFF)
            if data._ranks is None:
                raise SerializationError("ranks must be provided.")
            if len(data._ranks) != 9:
                raise SerializationError(f"Expected length of ranks to be exactly 9, got {len(data._ranks)}.")
            for i in range(9):
                writer.add_string(data._ranks[i])
                writer.add_byte(0xFF)
            if data._staff_count is None:
                raise SerializationError("staff_count must be provided.")
            writer.add_short(data._staff_count)
            writer.add_byte(0xFF)
            if data._staff is None:
                raise SerializationError("staff must be provided.")
            if len(data._staff) > 64008:
                raise SerializationError(f"Expected length of staff to be 64008 or less, got {len(data._staff)}.")
            for i in range(data._staff_count):
                GuildStaff.serialize(writer, data._staff[i])
                writer.add_byte(0xFF)
            writer.string_sanitization_mode = False
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "GuildReportServerPacket":
        """
        Deserializes an instance of `GuildReportServerPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            GuildReportServerPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            reader.chunked_reading_mode = True
            name = reader.get_string()
            reader.next_chunk()
            tag = reader.get_string()
            reader.next_chunk()
            create_date = reader.get_string()
            reader.next_chunk()
            description = reader.get_string()
            reader.next_chunk()
            wealth = reader.get_string()
            reader.next_chunk()
            ranks = []
            for i in range(9):
                ranks.append(reader.get_string())
                reader.next_chunk()
            staff_count = reader.get_short()
            reader.next_chunk()
            staff = []
            for i in range(staff_count):
                staff.append(GuildStaff.deserialize(reader))
                reader.next_chunk()
            reader.chunked_reading_mode = False
            result = GuildReportServerPacket(name=name, tag=tag, create_date=create_date, description=description, wealth=wealth, ranks=ranks, staff=staff)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"GuildReportServerPacket(byte_size={repr(self._byte_size)}, name={repr(self._name)}, tag={repr(self._tag)}, create_date={repr(self._create_date)}, description={repr(self._description)}, wealth={repr(self._wealth)}, ranks={repr(self._ranks)}, staff={repr(self._staff)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from ....serialization_error import SerializationError
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class PartyMember:
    """
    A member of the player's party
    """
    _byte_size: int = 0
    _player_id: int
    _leader: bool
    _level: int
    _hp_percentage: int
    _name: str

    def __init__(self, *, player_id: int, leader: bool, level: int, hp_percentage: int, name: str):
        """
        Create a new instance of PartyMember.

        Args:
            player_id (int): (Value range is 0-64008.)
            leader (bool): 
            level (int): (Value range is 0-252.)
            hp_percentage (int): (Value range is 0-252.)
            name (str): 
        """
        self._player_id = player_id
        self._leader = leader
        self._level = level
        self._hp_percentage = hp_percentage
        self._name = name

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def player_id(self) -> int:
        return self._player_id

    @property
    def leader(self) -> bool:
        return self._leader

    @property
    def level(self) -> int:
        return self._level

    @property
    def hp_percentage(self) -> int:
        return self._hp_percentage

    @property
    def name(self) -> str:
        return self._name

    @staticmethod
    def serialize(writer: EoWriter, data: "PartyMember") -> None:
        """
        Serializes an instance of `PartyMember` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (PartyMember): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._player_id is None:
                raise SerializationError("player_id must be provided.")
            writer.add_short(data._player_id)
            if data._leader is None:
                raise SerializationError("leader must be provided.")
            writer.add_char(1 if data._leader else 0)
            if data._level is None:
                raise SerializationError("level must be provided.")
            writer.add_char(data._level)
            if data._hp_percentage is None:
                raise SerializationError("hp_percentage must be provided.")
            writer.add_char(data._hp_percentage)
            if data._name is None:
                raise SerializationError("name must be provided.")
            writer.add_string(data._name)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "PartyMember":
        """
        Deserializes an instance of `PartyMember` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            PartyMember: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            player_id = reader.get_short()
            leader = reader.get_char() != 0
            level = reader.get_char()
            hp_percentage = reader.get_char()
            name = reader.get_string()
            result = PartyMember(player_id=player_id, leader=leader, level=level, hp_percentage=hp_percentage, name=name)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"PartyMember(byte_size={repr(self._byte_size)}, player_id={repr(self._player_id)}, leader={repr(self._leader)}, level={repr(self._level)}, hp_percentage={repr(self._hp_percentage)}, name={repr(self._name)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ....protocol_enum_meta import ProtocolEnumMeta

class SitState(IntEnum, metaclass=ProtocolEnumMeta):
    """
    Indicates how a player is sitting (or not sitting)
    """
    Stand = 0
    Chair = 1
    Floor = 2



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from __future__ import annotations
from collections.abc import Iterable
from .inn_record import InnRecord
from ....serialization_error import SerializationError
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class InnFile:
    """
    Endless Inn File
    """
    _byte_size: int = 0
    _inns: tuple[InnRecord, ...]

    def __init__(self, *, inns: Iterable[InnRecord]):
        """
        Create a new instance of InnFile.

        Args:
            inns (Iterable[InnRecord]): 
        """
        self._inns = tuple(inns)

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def inns(self) -> tuple[InnRecord, ...]:
        return self._inns

    @staticmethod
    def serialize(writer: EoWriter, data: "InnFile") -> None:
        """
        Serializes an instance of `InnFile` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (InnFile): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            writer.add_fixed_string("EID", 3, False)
            if data._inns is None:
                raise SerializationError("inns must be provided.")
            for i in range(len(data._inns)):
                InnRecord.serialize(writer, data._inns[i])
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "InnFile":
        """
        Deserializes an instance of `InnFile` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            InnFile: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            reader.get_fixed_string(3, False)
            inns = []
            while reader.remaining > 0:
                inns.append(InnRecord.deserialize(reader))
            result = InnFile(inns=inns)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"InnFile(byte_size={repr(self._byte_size)}, inns={repr(self._inns)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ....protocol_enum_meta import ProtocolEnumMeta

class WarpType(IntEnum, metaclass=ProtocolEnumMeta):
    """
    Indicates whether a warp is within the current map, or switching to another map
    """
    Local = 1
    MapSwitch = 2



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from ....serialization_error import SerializationError
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class CharacterElementalStats:
    """
    The 6 elemental character stats
    """
    _byte_size: int = 0
    _light: int
    _dark: int
    _fire: int
    _water: int
    _earth: int
    _wind: int

    def __init__(self, *, light: int, dark: int, fire: int, water: int, earth: int, wind: int):
        """
        Create a new instance of CharacterElementalStats.

        Args:
            light (int): (Value range is 0-64008.)
            dark (int): (Value range is 0-64008.)
            fire (int): (Value range is 0-64008.)
            water (int): (Value range is 0-64008.)
            earth (int): (Value range is 0-64008.)
            wind (int): (Value range is 0-64008.)
        """
        self._light = light
        self._dark = dark
        self._fire = fire
        self._water = water
        self._earth = earth
        self._wind = wind

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def light(self) -> int:
        return self._light

    @property
    def dark(self) -> int:
        return self._dark

    @property
    def fire(self) -> int:
        return self._fire

    @property
    def water(self) -> int:
        return self._water

    @property
    def earth(self) -> int:
        return self._earth

    @property
    def wind(self) -> int:
        return self._wind

    @staticmethod
    def serialize(writer: EoWriter, data: "CharacterElementalStats") -> None:
        """
        Serializes an instance of `CharacterElementalStats` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (CharacterElementalStats): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._light is None:
                raise SerializationError("light must be provided.")
            writer.add_short(data._light)
            if data._dark is None:
                raise SerializationError("dark must be provided.")
            writer.add_short(data._dark)
            if data._fire is None:
                raise SerializationError("fire must be provided.")
            writer.add_short(data._fire)
            if data._water is None:
                raise SerializationError("water must be provided.")
            writer.add_short(data._water)
            if data._earth is None:
                raise SerializationError("earth must be provided.")
            writer.add_short(data._earth)
            if data._wind is None:
                raise SerializationError("wind must be provided.")
            writer.add_short(data._wind)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "CharacterElementalStats":
        """
        Deserializes an instance of `CharacterElementalStats` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            CharacterElementalStats: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            light = reader.get_short()
            dark = reader.get_short()
            fire = reader.get_short()
            water = reader.get_short()
            earth = reader.get_short()
            wind = reader.get_short()
            result = CharacterElementalStats(light=light, dark=dark, fire=fire, water=water, earth=earth, wind=wind)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"CharacterElementalStats(byte_size={repr(self._byte_size)}, light={repr(self._light)}, dark={repr(self._dark)}, fire={repr(self._fire)}, water={repr(self._water)}, earth={repr(self._earth)}, wind={repr(self._wind)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from ...serialization_error import SerializationError
from ....data.eo_writer import EoWriter
from ....data.eo_reader import EoReader

class Weight:
    """
    Current carry weight and maximum carry capacity of a player
    """
    _byte_size: int = 0
    _current: int
    _max: int

    def __init__(self, *, current: int, max: int):
        """
        Create a new instance of Weight.

        Args:
            current (int): (Value range is 0-252.)
            max (int): (Value range is 0-252.)
        """
        self._current = current
        self._max = max

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def current(self) -> int:
        return self._current

    @property
    def max(self) -> int:
        return self._max

    @staticmethod
    def serialize(writer: EoWriter, data: "Weight") -> None:
        """
        Serializes an instance of `Weight` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (Weight): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._current is None:
                raise SerializationError("current must be provided.")
            writer.add_char(data._current)
            if data._max is None:
                raise SerializationError("max must be provided.")
            writer.add_char(data._max)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "Weight":
        """
        Deserializes an instance of `Weight` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            Weight: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            current = reader.get_char()
            max = reader.get_char()
            result = Weight(current=current, max=max)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"Weight(byte_size={repr(self._byte_size)}, current={repr(self._current)}, max={repr(self._max)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ....protocol_enum_meta import ProtocolEnumMeta

class LoginMessageCode(IntEnum, metaclass=ProtocolEnumMeta):
    """
    Whether a warning message should be displayed upon entering the game
    """
    No = 0
    Yes = 2



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from ..party_request_type import PartyRequestType
from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class PartyAcceptClientPacket(Packet):
    """
    Accept party invite / join request
    """
    _byte_size: int = 0
    _request_type: PartyRequestType
    _inviter_player_id: int

    def __init__(self, *, request_type: PartyRequestType, inviter_player_id: int):
        """
        Create a new instance of PartyAcceptClientPacket.

        Args:
            request_type (PartyRequestType): 
            inviter_player_id (int): (Value range is 0-64008.)
        """
        self._request_type = request_type
        self._inviter_player_id = inviter_player_id

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def request_type(self) -> PartyRequestType:
        return self._request_type

    @property
    def inviter_player_id(self) -> int:
        return self._inviter_player_id

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.Party

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Accept

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        PartyAcceptClientPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "PartyAcceptClientPacket") -> None:
        """
        Serializes an instance of `PartyAcceptClientPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (PartyAcceptClientPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._request_type is None:
                raise SerializationError("request_type must be provided.")
            writer.add_char(int(data._request_type))
            if data._inviter_player_id is None:
                raise SerializationError("inviter_player_id must be provided.")
            writer.add_short(data._inviter_player_id)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "PartyAcceptClientPacket":
        """
        Deserializes an instance of `PartyAcceptClientPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            PartyAcceptClientPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            request_type = PartyRequestType(reader.get_char())
            inviter_player_id = reader.get_short()
            result = PartyAcceptClientPacket(request_type=request_type, inviter_player_id=inviter_player_id)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"PartyAcceptClientPacket(byte_size={repr(self._byte_size)}, request_type={repr(self._request_type)}, inviter_player_id={repr(self._inviter_player_id)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ....protocol_enum_meta import ProtocolEnumMeta

class FileType(IntEnum, metaclass=ProtocolEnumMeta):
    """
    Data file type
    """
    Emf = 1
    Eif = 2
    Enf = 3
    Esf = 4
    Ecf = 5



from enum import IntEnum
from eolib import ProtocolEnumMeta


class ProtocolEnum(IntEnum, metaclass=ProtocolEnumMeta):
    FOO = 0
    BAR = 1
    BAZ = 2


def test_known_value():
    assert ProtocolEnum(0) is ProtocolEnum.FOO
    assert ProtocolEnum(1) is ProtocolEnum.BAR
    assert ProtocolEnum(2) is ProtocolEnum.BAZ


def test_unrecognized_value():
    unrecognized = ProtocolEnum(3)
    assert isinstance(unrecognized, ProtocolEnum)
    assert unrecognized.name == "Unrecognized(3)"
    assert unrecognized.value == 3



from protocol_code_generator.util.number_utils import try_parse_int


class Length:
    def __init__(self, length_string):
        self._string = length_string
        self._integer = try_parse_int(length_string)

    @classmethod
    def from_string(cls, length_string):
        return cls(length_string)

    @classmethod
    def unspecified(cls):
        return cls(None)

    def as_integer(self):
        return self._integer

    @property
    def specified(self):
        return self._string is not None

    def __str__(self):
        return self._string if self.specified else "[unspecified]"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from typing import cast
from typing import Optional
from .npc_kill_steal_protection_state import NpcKillStealProtectionState
from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ...direction import Direction
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class CastReplyServerPacket(Packet):
    """
    Nearby NPC hit by a spell from a player
    """
    _byte_size: int = 0
    _spell_id: int
    _caster_id: int
    _caster_direction: Direction
    _npc_index: int
    _damage: int
    _hp_percentage: int
    _caster_tp: Optional[int]
    _kill_steal_protection: Optional[NpcKillStealProtectionState]

    def __init__(self, *, spell_id: int, caster_id: int, caster_direction: Direction, npc_index: int, damage: int, hp_percentage: int, caster_tp: Optional[int] = None, kill_steal_protection: Optional[NpcKillStealProtectionState] = None):
        """
        Create a new instance of CastReplyServerPacket.

        Args:
            spell_id (int): (Value range is 0-64008.)
            caster_id (int): (Value range is 0-64008.)
            caster_direction (Direction): 
            npc_index (int): (Value range is 0-64008.)
            damage (int): (Value range is 0-16194276.)
            hp_percentage (int): (Value range is 0-64008.)
            caster_tp (Optional[int]): This field should be sent to the attacker, but not nearby players (Value range is 0-64008.)
            kill_steal_protection (Optional[NpcKillStealProtectionState]): This field should be sent to the attacker, but not nearby players
        """
        self._spell_id = spell_id
        self._caster_id = caster_id
        self._caster_direction = caster_direction
        self._npc_index = npc_index
        self._damage = damage
        self._hp_percentage = hp_percentage
        self._caster_tp = caster_tp
        self._kill_steal_protection = kill_steal_protection

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def spell_id(self) -> int:
        return self._spell_id

    @property
    def caster_id(self) -> int:
        return self._caster_id

    @property
    def caster_direction(self) -> Direction:
        return self._caster_direction

    @property
    def npc_index(self) -> int:
        return self._npc_index

    @property
    def damage(self) -> int:
        return self._damage

    @property
    def hp_percentage(self) -> int:
        return self._hp_percentage

    @property
    def caster_tp(self) -> Optional[int]:
        """
        This field should be sent to the attacker, but not nearby players
        """
        return self._caster_tp

    @property
    def kill_steal_protection(self) -> Optional[NpcKillStealProtectionState]:
        """
        This field should be sent to the attacker, but not nearby players
        """
        return self._kill_steal_protection

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.Cast

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Reply

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        CastReplyServerPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "CastReplyServerPacket") -> None:
        """
        Serializes an instance of `CastReplyServerPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (CastReplyServerPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._spell_id is None:
                raise SerializationError("spell_id must be provided.")
            writer.add_short(data._spell_id)
            if data._caster_id is None:
                raise SerializationError("caster_id must be provided.")
            writer.add_short(data._caster_id)
            if data._caster_direction is None:
                raise SerializationError("caster_direction must be provided.")
            writer.add_char(int(data._caster_direction))
            if data._npc_index is None:
                raise SerializationError("npc_index must be provided.")
            writer.add_short(data._npc_index)
            if data._damage is None:
                raise SerializationError("damage must be provided.")
            writer.add_three(data._damage)
            if data._hp_percentage is None:
                raise SerializationError("hp_percentage must be provided.")
            writer.add_short(data._hp_percentage)
            reached_missing_optional = data._caster_tp is None
            if not reached_missing_optional:
                writer.add_short(cast(int, data._caster_tp))
            reached_missing_optional = reached_missing_optional or data._kill_steal_protection is None
            if not reached_missing_optional:
                writer.add_char(int(cast(NpcKillStealProtectionState, data._kill_steal_protection)))
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "CastReplyServerPacket":
        """
        Deserializes an instance of `CastReplyServerPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            CastReplyServerPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            spell_id = reader.get_short()
            caster_id = reader.get_short()
            caster_direction = Direction(reader.get_char())
            npc_index = reader.get_short()
            damage = reader.get_three()
            hp_percentage = reader.get_short()
            caster_tp: Optional[int] = None
            if reader.remaining > 0:
                caster_tp = reader.get_short()
            kill_steal_protection: Optional[NpcKillStealProtectionState] = None
            if reader.remaining > 0:
                kill_steal_protection = NpcKillStealProtectionState(reader.get_char())
            result = CastReplyServerPacket(spell_id=spell_id, caster_id=caster_id, caster_direction=caster_direction, npc_index=npc_index, damage=damage, hp_percentage=hp_percentage, caster_tp=caster_tp, kill_steal_protection=kill_steal_protection)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"CastReplyServerPacket(byte_size={repr(self._byte_size)}, spell_id={repr(self._spell_id)}, caster_id={repr(self._caster_id)}, caster_direction={repr(self._caster_direction)}, npc_index={repr(self._npc_index)}, damage={repr(self._damage)}, hp_percentage={repr(self._hp_percentage)}, caster_tp={repr(self._caster_tp)}, kill_steal_protection={repr(self._kill_steal_protection)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ...protocol_enum_meta import ProtocolEnumMeta

class NpcType(IntEnum, metaclass=ProtocolEnumMeta):
    Friendly = 0
    Passive = 1
    Aggressive = 2
    Reserved3 = 3
    Reserved4 = 4
    Reserved5 = 5
    Shop = 6
    Inn = 7
    Reserved8 = 8
    Bank = 9
    Barber = 10
    Guild = 11
    Priest = 12
    Lawyer = 13
    Trainer = 14
    Quest = 15



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from __future__ import annotations
from collections.abc import Iterable
from ..three_item import ThreeItem
from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ..item import Item
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class AdminInteractListServerPacket(Packet):
    """
    Admin character inventory popup
    """
    _byte_size: int = 0
    _name: str
    _usage: int
    _gold_bank: int
    _inventory: tuple[Item, ...]
    _bank: tuple[ThreeItem, ...]

    def __init__(self, *, name: str, usage: int, gold_bank: int, inventory: Iterable[Item], bank: Iterable[ThreeItem]):
        """
        Create a new instance of AdminInteractListServerPacket.

        Args:
            name (str): 
            usage (int): (Value range is 0-4097152080.)
            gold_bank (int): (Value range is 0-4097152080.)
            inventory (Iterable[Item]): 
            bank (Iterable[ThreeItem]): 
        """
        self._name = name
        self._usage = usage
        self._gold_bank = gold_bank
        self._inventory = tuple(inventory)
        self._bank = tuple(bank)

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def name(self) -> str:
        return self._name

    @property
    def usage(self) -> int:
        return self._usage

    @property
    def gold_bank(self) -> int:
        return self._gold_bank

    @property
    def inventory(self) -> tuple[Item, ...]:
        return self._inventory

    @property
    def bank(self) -> tuple[ThreeItem, ...]:
        return self._bank

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.AdminInteract

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.List

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        AdminInteractListServerPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "AdminInteractListServerPacket") -> None:
        """
        Serializes an instance of `AdminInteractListServerPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (AdminInteractListServerPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            writer.string_sanitization_mode = True
            if data._name is None:
                raise SerializationError("name must be provided.")
            writer.add_string(data._name)
            writer.add_byte(0xFF)
            if data._usage is None:
                raise SerializationError("usage must be provided.")
            writer.add_int(data._usage)
            writer.add_byte(0xFF)
            if data._gold_bank is None:
                raise SerializationError("gold_bank must be provided.")
            writer.add_int(data._gold_bank)
            writer.add_byte(0xFF)
            if data._inventory is None:
                raise SerializationError("inventory must be provided.")
            for i in range(len(data._inventory)):
                Item.serialize(writer, data._inventory[i])
            writer.add_byte(0xFF)
            if data._bank is None:
                raise SerializationError("bank must be provided.")
            for i in range(len(data._bank)):
                ThreeItem.serialize(writer, data._bank[i])
            writer.string_sanitization_mode = False
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "AdminInteractListServerPacket":
        """
        Deserializes an instance of `AdminInteractListServerPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            AdminInteractListServerPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            reader.chunked_reading_mode = True
            name = reader.get_string()
            reader.next_chunk()
            usage = reader.get_int()
            reader.next_chunk()
            gold_bank = reader.get_int()
            reader.next_chunk()
            inventory_length = int(reader.remaining / 6)
            inventory = []
            for i in range(inventory_length):
                inventory.append(Item.deserialize(reader))
            reader.next_chunk()
            bank_length = int(reader.remaining / 5)
            bank = []
            for i in range(bank_length):
                bank.append(ThreeItem.deserialize(reader))
            reader.chunked_reading_mode = False
            result = AdminInteractListServerPacket(name=name, usage=usage, gold_bank=gold_bank, inventory=inventory, bank=bank)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"AdminInteractListServerPacket(byte_size={repr(self._byte_size)}, name={repr(self._name)}, usage={repr(self._usage)}, gold_bank={repr(self._gold_bank)}, inventory={repr(self._inventory)}, bank={repr(self._bank)})"



from eolib.data.eo_numeric_limits import CHAR_MAX, SHORT_MAX, THREE_MAX, INT_MAX
from eolib.data.number_encoding_utils import encode_number
from eolib.data.string_encoding_utils import encode_string


class EoWriter:
    def __init__(self):
        self.data = bytearray()
        self._string_sanitization_mode = False

    def add_byte(self, value: int) -> None:
        """
        Adds a raw byte to the writer data.

        Args:
            value (int): The byte to add.

        Raises:
            ValueError: If the value is above `0xFF`.
        """
        self._check_number_size(value, 0xFF)
        self.data.append(value)

    def add_bytes(self, bytes: bytes) -> None:
        """
        Adds raw bytes to the writer data.

        Args:
            bytes (bytes): The bytes to add.
        """
        self.data.extend(bytes)

    def add_char(self, number: int) -> None:
        """
        Adds an encoded 1-byte integer to the writer data.

        Args:
            number (int): The number to encode and add.

        Raises:
            ValueError: If the value is not below `CHAR_MAX`.
        """
        self._check_number_size(number, CHAR_MAX - 1)
        number_bytes = encode_number(number)
        self._add_bytes_with_length(number_bytes, 1)

    def add_short(self, number: int) -> None:
        """
        Adds an encoded 2-byte integer to the writer data.

        Args:
            number (int): The number to encode and add.

        Raises:
            ValueError: If the value is not below `SHORT_MAX`.
        """
        self._check_number_size(number, SHORT_MAX - 1)
        number_bytes = encode_number(number)
        self._add_bytes_with_length(number_bytes, 2)

    def add_three(self, number: int) -> None:
        """
        Adds an encoded 3-byte integer to the writer data.

        Args:
            number (int): The number to encode and add.

        Raises:
            ValueError: If the value is not below `THREE_MAX`.
        """
        self._check_number_size(number, THREE_MAX - 1)
        number_bytes = encode_number(number)
        self._add_bytes_with_length(number_bytes, 3)

    def add_int(self, number: int) -> None:
        """
        Adds an encoded 4-byte integer to the writer data.

        Args:
            number (int): The number to encode and add.

        Raises:
            ValueError: If the value is not below `INT_MAX`.
        """
        self._check_number_size(number, INT_MAX - 1)
        number_bytes = encode_number(number)
        self._add_bytes_with_length(number_bytes, 4)

    def add_string(self, string: str) -> None:
        """
        Adds a string to the writer data.

        Args:
            string (str): The string to be added.
        """
        string_bytes = self._encode_ansi(string)
        self._sanitize_string(string_bytes)
        self.add_bytes(string_bytes)

    def add_fixed_string(self, string: str, length: int, padded: bool = False) -> None:
        """
        Adds a fixed-length string to the writer data.

        Args:
            string (str): The string to be added.
            length (int): The expected length of the string.
            padded (bool, optional): True if the string should be padded to the length with
                trailing `0xFF` bytes. Defaults to False.
        """
        self._check_string_length(string, length, padded)
        string_bytes = self._encode_ansi(string)
        self._sanitize_string(string_bytes)
        if padded:
            string_bytes = self._add_padding(string_bytes, length)
        self.add_bytes(string_bytes)

    def add_encoded_string(self, string: str) -> None:
        """
        Adds an encoded string to the writer data.

        Args:
            string (str): The string to be encoded and added.
        """
        string_bytes = self._encode_ansi(string)
        self._sanitize_string(string_bytes)
        encode_string(string_bytes)
        self.add_bytes(string_bytes)

    def add_fixed_encoded_string(self, string: str, length: int, padded: bool = False) -> None:
        """
        Adds a fixed-length encoded string to the writer data.

        Args:
            string (str): The string to be encoded and added.
            length (int): The expected length of the string.
            padded (bool, optional): True if the string should be padded to the length with
                trailing `0xFF` bytes. Defaults to False.
        """
        self._check_string_length(string, length, padded)
        string_bytes = self._encode_ansi(string)
        self._sanitize_string(string_bytes)
        if padded:
            string_bytes = self._add_padding(string_bytes, length)
        encode_string(string_bytes)
        self.add_bytes(string_bytes)

    @property
    def string_sanitization_mode(self) -> bool:
        """
        Gets the string sanitization mode for the writer.

        Returns:
            bool: True if string sanitization is enabled.
        """
        return self._string_sanitization_mode

    @string_sanitization_mode.setter
    def string_sanitization_mode(self, string_sanitization_mode: bool) -> None:
        self._string_sanitization_mode = string_sanitization_mode

    def to_bytearray(self) -> bytearray:
        """
        Gets the writer data as a byte array.

        Returns:
            bytearray: A copy of the writer data as a byte array.
        """
        return self.data.copy()

    def __len__(self) -> int:
        """
        Gets the length of the writer data.

        Returns:
            int: The length of the writer data.
        """
        return len(self.data)

    def _add_bytes_with_length(self, bytes: bytes, bytes_length: int) -> None:
        """
        Adds raw bytes with a specified length to the writer data.

        Args:
            bytes (bytes): The bytes to add.
            bytes_length (int): The number of bytes to add.
        """
        self.data.extend(bytes[:bytes_length])

    def _sanitize_string(self, bytes: bytearray) -> None:
        if self.string_sanitization_mode:
            for i in range(len(bytes)):
                if bytes[i] == 0xFF:  # 'ÿ'
                    bytes[i] = 0x79  # 'y'

    @staticmethod
    def _check_number_size(number: int, max_value: int) -> None:
        if number > max_value:
            raise ValueError(f"Value {number} exceeds maximum of {max_value}.")

    @staticmethod
    def _add_padding(bytes: bytearray, length: int) -> bytearray:
        if len(bytes) == length:
            return bytes

        result = bytearray(length)
        result[: len(bytes)] = bytes
        result[len(bytes) :] = bytearray([0xFF] * (length - len(bytes)))

        return result

    @staticmethod
    def _check_string_length(string: str, length: int, padded: bool) -> None:
        if padded:
            if length >= len(string):
                return
            raise ValueError(f'Padded string "{string}" is too large for a length of {length}.')

        if len(string) != length:
            raise ValueError(f'String "{string}" does not have expected length of {length}.')

    @staticmethod
    def _encode_ansi(string: str) -> bytearray:
        """
        Encodes string to windows-1252 bytes.

        Args:
            string (str): The string to encode.

        Returns:
            bytearray: The encoded string.
        """
        return bytearray(string, 'windows-1252', 'replace')



import copy

from protocol_code_generator.generate.code_block import CodeBlock
from protocol_code_generator.type.enum_type import EnumType
from protocol_code_generator.type.integer_type import IntegerType
from protocol_code_generator.util.docstring_utils import generate_docstring
from protocol_code_generator.util.name_utils import snake_case_to_pascal_case
from protocol_code_generator.util.number_utils import try_parse_int
from protocol_code_generator.util.xml_utils import (
    get_boolean_attribute,
    get_comment,
    get_instructions,
    get_required_string_attribute,
)


class SwitchCodeGenerator:
    def __init__(self, field_name, type_factory, context, data):
        self._field_name = field_name
        self._type_factory = type_factory
        self._context = context
        self._data = data

    def generate_case_data_interface(self, protocol_cases):
        union_type_names = [
            f"'{self.get_case_data_type_name(case)}'"
            for case in protocol_cases
            if len(get_instructions(case)) > 0
        ]
        union_type_names.append("None")

        union_type = f"Union[{', '.join(union_type_names)}]"
        field_name = self._field_name

        self._data.add_auxiliary_type(
            CodeBlock()
            .add_line(f"{self._interface_type_name} = {union_type}")
            .add_line('"""')
            .add_line(f'Data associated with different values of the `{field_name}` field.')
            .add_line('"""')
            .add_import("Union", "typing")
        )

    def generate_case_data_field(self):
        interface_type_name = f"{self._data.class_name}.{self._interface_type_name}"
        case_data_field_name = self._case_data_field_name
        switch_field_name = self._field_name

        self._data.fields.add_line(f"_{case_data_field_name}: '{interface_type_name}'")
        self._data.init_docstring_params.append(
            CodeBlock().add(
                f'{case_data_field_name} ({interface_type_name}): Data associated with the '
                f'`{switch_field_name}` field.'
            )
        )
        self._data.add_method(
            CodeBlock()
            .add_line("@property")
            .add_line(f"def {case_data_field_name}(self) -> '{interface_type_name}':")
            .indent()
            .add_line('"""')
            .add_line(
                f'{interface_type_name}: Data associated with the `{switch_field_name}` field.'
            )
            .add_line('"""')
            .add_line(f"return self._{case_data_field_name}")
            .unindent()
        )
        self._data.repr_fields.append(case_data_field_name)
        self._data.init_params.append(
            CodeBlock().add(f"{case_data_field_name}: '{interface_type_name}' = None")
        )
        self._data.init_body.add_line(f'self._{case_data_field_name} = {case_data_field_name}')
        self._data.deserialize.add_line(f'{case_data_field_name}: {interface_type_name} = None')
        self._data.deserialize_init_arguments.append(
            CodeBlock().add(f'{case_data_field_name}={case_data_field_name}')
        )

    def generate_case(self, protocol_case, start):
        case_data_type_name = self.get_case_data_type_name(protocol_case)
        case_context = copy.deepcopy(self._context)
        case_context.accessible_fields.clear()
        case_context.length_field_is_referenced_map.clear()

        default = get_boolean_attribute(protocol_case, "default")

        if default:
            if start:
                raise RuntimeError("Standalone default case is not allowed.")
            self._data.serialize.begin_control_flow('else')
            self._data.deserialize.begin_control_flow('else')
        else:
            keyword = 'if' if start else 'elif'
            case_value_expression = self._get_case_value_expression(protocol_case)
            self._data.serialize.begin_control_flow(
                f'{keyword} data._{self._field_name} == {case_value_expression}'
            )
            self._data.deserialize.begin_control_flow(
                f'{keyword} {self._field_name} == {case_value_expression}'
            )

        field_to_string_expression = (
            f"{self._field_data.type_.name}(data._{self._field_name}).name"
            if isinstance(self._field_data.type_, EnumType)
            else f"str(data._{self._field_name})"
        )

        if get_instructions(protocol_case) == []:
            self._data.serialize.begin_control_flow(
                f"if data._{self._case_data_field_name} is not None"
            )
            self._data.serialize.add_line(
                'raise SerializationError('
                + f'"Expected {self._case_data_field_name} to be None for {self._field_name} "'
                + f' + {field_to_string_expression} + ".")'
            )
            self._data.serialize.unindent()
            self._data.serialize.add_import(
                "SerializationError", "eolib.protocol.serialization_error"
            )

            self._data.deserialize.add_line(f"{self._case_data_field_name} = None")

        else:
            self._data.add_auxiliary_type(
                self.generate_case_data_type(protocol_case, case_data_type_name, case_context)
            )
            self._data.serialize.begin_control_flow(
                f"if not isinstance(data._{self._case_data_field_name}, {case_data_type_name})"
            )
            self._data.serialize.add_line(
                'raise SerializationError('
                + f'"Expected {self._case_data_field_name} to be type {case_data_type_name} '
                + f'for {self._field_name} " + {field_to_string_expression} + ".")'
            )
            self._data.serialize.unindent()
            self._data.serialize.add_line(
                f'{case_data_type_name}.serialize(writer, data._{self._case_data_field_name})'
            )
            self._data.serialize.add_import(
                "SerializationError", "eolib.protocol.serialization_error"
            )

            self._data.deserialize.add_line(
                f'{self._case_data_field_name} = {case_data_type_name}.deserialize(reader)'
            )

        self._data.serialize.unindent()
        self._data.deserialize.unindent()

        return case_context

    def generate_case_data_type(self, protocol_case, case_data_type_name, case_context):
        from protocol_code_generator.generate.object_code_generator import ObjectCodeGenerator

        object_code_generator = ObjectCodeGenerator(
            case_data_type_name, self._type_factory, case_context
        )

        for instruction in get_instructions(protocol_case):
            object_code_generator.generate_instruction(instruction)

        default = get_boolean_attribute(protocol_case, "default")

        if default:
            comment = f'Default data associated with {self._field_name}'
        else:
            case_value_expression = self._get_case_value_expression(protocol_case)
            comment = f'Data associated with {self._field_name} value {case_value_expression}'

        protocol_comment = get_comment(protocol_case)
        if protocol_comment is not None:
            comment += '\n\n'
            comment += protocol_comment

        object_code_generator.data.docstring = generate_docstring(comment)

        return object_code_generator.code

    @property
    def _field_data(self):
        result = self._context.accessible_fields.get(self._field_name)
        if not result:
            raise RuntimeError(f"Referenced {self._field_name} is not accessible.")
        return result

    @property
    def _interface_type_name(self):
        return snake_case_to_pascal_case(self._field_name) + "Data"

    @property
    def _case_data_field_name(self):
        return self._field_name + "_data"

    def get_case_data_type_name(self, protocol_case):
        is_default = get_boolean_attribute(protocol_case, "default")
        return (
            self._data.class_name
            + '.'
            + self._interface_type_name
            + ("Default" if is_default else get_required_string_attribute(protocol_case, "value"))
        )

    def _get_case_value_expression(self, protocol_case):
        field_data = self._field_data

        if field_data.array:
            raise RuntimeError(
                f'"{self._field_name}" field referenced by switch must not be an array.'
            )

        field_type = field_data.type_
        case_value = get_required_string_attribute(protocol_case, "value")

        if isinstance(field_type, IntegerType):
            if not case_value.isdigit():
                raise RuntimeError(f'"{case_value}" is not a valid integer value.')
            return case_value

        if isinstance(field_type, EnumType):
            ordinal_value = try_parse_int(case_value)
            if ordinal_value is not None:
                enum_value = field_type.get_enum_value_by_ordinal(ordinal_value)
                if enum_value is not None:
                    raise RuntimeError(
                        f'{field_type.name} value {case_value} '
                        + f'must be referred to by name ({enum_value.name})'
                    )
                return case_value

            enum_value = field_type.get_enum_value_by_name(case_value)
            if enum_value is None:
                raise RuntimeError(
                    f'"{case_value}" is not a valid value for enum type {field_type.name}.'
                )
            return f'{field_type.name}.{enum_value.python_name}'

        raise RuntimeError(
            f'{self._field_name} field referenced by switch must be a numeric or enumeration type.'
        )



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from __future__ import annotations
from collections.abc import Iterable
from .map_graphic_row import MapGraphicRow
from ...serialization_error import SerializationError
from ....data.eo_writer import EoWriter
from ....data.eo_reader import EoReader

class MapGraphicLayer:
    """
    A layer of map graphics
    """
    _byte_size: int = 0
    _graphic_rows_count: int
    _graphic_rows: tuple[MapGraphicRow, ...]

    def __init__(self, *, graphic_rows: Iterable[MapGraphicRow]):
        """
        Create a new instance of MapGraphicLayer.

        Args:
            graphic_rows (Iterable[MapGraphicRow]): (Length must be 252 or less.)
        """
        self._graphic_rows = tuple(graphic_rows)
        self._graphic_rows_count = len(self._graphic_rows)

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def graphic_rows(self) -> tuple[MapGraphicRow, ...]:
        return self._graphic_rows

    @staticmethod
    def serialize(writer: EoWriter, data: "MapGraphicLayer") -> None:
        """
        Serializes an instance of `MapGraphicLayer` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (MapGraphicLayer): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._graphic_rows_count is None:
                raise SerializationError("graphic_rows_count must be provided.")
            writer.add_char(data._graphic_rows_count)
            if data._graphic_rows is None:
                raise SerializationError("graphic_rows must be provided.")
            if len(data._graphic_rows) > 252:
                raise SerializationError(f"Expected length of graphic_rows to be 252 or less, got {len(data._graphic_rows)}.")
            for i in range(data._graphic_rows_count):
                MapGraphicRow.serialize(writer, data._graphic_rows[i])
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "MapGraphicLayer":
        """
        Deserializes an instance of `MapGraphicLayer` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            MapGraphicLayer: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            graphic_rows_count = reader.get_char()
            graphic_rows = []
            for i in range(graphic_rows_count):
                graphic_rows.append(MapGraphicRow.deserialize(reader))
            result = MapGraphicLayer(graphic_rows=graphic_rows)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"MapGraphicLayer(byte_size={repr(self._byte_size)}, graphic_rows={repr(self._graphic_rows)})"
